/*
 Copyright (C) 2018 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief  This macro imports a Federeated Area (FA) into the user's PFA (Personal Federated Area) on the SAS Server. The PFA is made available for use without needing a server restart.
           The macro may be executed in an interactive session such as SAS Studio or a background process.

   <b> Identified Inputs </b> 

   \param [in] meta_host               Complete name of the server running the SAS Metadata Server. 
   \param [in] meta_port               Port on which the SAS Metadats Server is listening. Defaults to 8561.
   \param [in] meta_repos              Name of the metadata repository. Defaults to Foundation.
   \param [in] username                Username credentials for logging into IRM (not required if logged-in to SAS Studio).
   \param [in] passwd                  Password credentials: it can be plain text or SAS-Encoded (will be masked during execution). Not required if logged-in to SAS Studio.
   \param [in] fa_src_path             Complete path (including directory name/zip file name) of the source location where the FA directory/zip exists.
   \param [in] force_flg   [OPTIONAL]  Flag (True/False) forcing/canceling install respectively if dependencies are not met or FA is invalid. Defaults to false.
   \param [in] debug       [OPTIONAL]  Amount of execution information in the log. Defaults to FALSE.
   \param [in] logOptions  [OPTIONAL]  Amount of macro execution information in the log. Defaults to blank. Accepts a space separated list of supported SAS options.
                                       Example: logOptions = mprint mlogic symbolgen
   
   \details 
            PURPOSE: The macro provides an interface to import a Federated Area into the user's PFA and make the PFA available for immediate use (no server restart).
            
            PRE-REQUISITES:
                     1. The FA to be imported is available at an accessible location on the server.
 
            USAGE:   Sample values have been provided for some parameters to illustrate usage.
                     %irm_import_fa(
                              meta_host=<host; metadata server>
                           ,  meta_port=<8561; optional for default metadata server port>
                           ,  meta_repos=<Foundation>
                           ,  username=<userid; optional in SAS Studio>
                           ,  passwd=<pwd; optional in SAS Studio>
                           ,  fa_src_path=<FA Src Path>
                           ,  force_flg=<true; optional>
                           ,  debug=<true; optional>
                           ,  logOptions=<mprint symbolgen; optional>
                     );

            OUTPUT:
                     1. Success: Federated Area imported into PFA and available for use
                     2. Failure: Appropriate messages in the execution log
                           
   \ingroup api
   
   \author  SAS Institute Inc.
   \date    2018
*/

%macro irm_import_fa(
         meta_host      =
      ,  meta_port      = 8561
      ,  meta_repos     = Foundation
      ,  username       =
      ,  passwd         =
      ,  fa_src_path    = 
      ,  force_flg      = false
      ,  debug          = false
      ,  logOptions     =
);
                                   
   %put ENTERING &sysmacroname.; 
   
   /* Declare LOCAL macro variables */
   %local
      SWCNAME
      server
      oldLogOptions
      suppressLogOptions      
   ; 
   
   /* Set required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
      
   /* Get current values of source, mlogic, mprint and symbolgen options */
   %let oldLogOptions = %sysfunc(getoption(source)) %sysfunc(getoption(mlogic)) %sysfunc(getoption(mprint)) %sysfunc(getoption(symbolgen));
   
   /* Set macro variable for suppressing log options */
   %let suppressLogOptions = nosource nomlogic nomprint nosymbolgen;

   /* Initialize */
   %let SWCNAME=IRM Mid-Tier Server;
   %let server = SASIRMServer;
   
   /* Check for valid inputs */
   %if (%bquote(&meta_host.) eq %str()) %then %do;
      %rsk_terminate(key=rsk_meta_host_missing);
      %return;
   %end;
   
   %if (&meta_port. eq %str()) %then
      %let meta_port=8561;
      
   %if (%bquote(&meta_repos.) eq %str()) %then
      %let meta_repos=Foundation;
   
   %if (%bquote(&fa_src_path.) eq %str()) %then %do;
      %rsk_terminate(key=rsk_fa_src_path_missing);
      %return;
   %end;
   
   /* Check that the source file location is valid */
   %if %sysfunc(fileexist(&fa_src_path.)) eq 0 %then
      %rsk_terminate(key=rsk_dir_not_found,s1="&fa_src_path.");   
   
   /* Check for username. If empty, could be a SAS Studio session. */
   %if (%bquote(&username.) eq %str()) %then %do;
      /* Assume a SAS Studio session and try to extract username. */
      %let username=&_CLIENTUSERNAME.;
      
      /* Still empty, then exit. */
      %if (%bquote(&username.) eq %str()) %then %do;
         %rsk_terminate(key=rsk_username_missing);
         %return;
      %end;
   %end;

   /* Query metadata to return mid-tier information. This will be used to construct REST URL:
      --host name
      --port number
      --protocol (http/https)
   */
   
   %global irm_host_name;
   %global irm_host_port;
   %global irm_host_proto;
   
   /* Suppress printing of passwords. */
   option &suppressLogOptions.;
   
   %irm_server_meta_query(
               i_meta_host    = &meta_host.
            ,  i_meta_port    = &meta_port.
            ,  i_meta_repos   = &meta_repos.
            ,  i_meta_user    = &username.
            ,  i_meta_passwd  = &passwd.
            ,  i_irm_server   = &server.            
            ,  o_host_name    = irm_host_name
            ,  o_host_port    = irm_host_port
            ,  o_host_proto   = irm_host_proto
   );
   
   /* Restore log options. */
   options &oldLogOptions.;
         
   /* Get Service ticket to PUT FA information through REST call */
   /* Constant String*/
   %local import_fa_url;
   %let import_fa_url = %str(/rest/federatedareas/import/);

   /* Define server URL */
   %local server_url;
   %let server_url = %str(&irm_host_proto.://&irm_host_name.:&irm_host_port./&server.);
   
   /* Define REST URL */
   %local rest_url;
   %let rest_url = %str(&server_url.&import_fa_url.);

   /* Check whether session is already authenticated. */
   %local auth_session_flg;
   %let auth_session_flg = N;
   %if %sysevalf("%superq(passwd)"="", boolean) %then
        %let auth_session_flg = Y;

   /* Get a Service ticket if credentials have been provided */
   %if(&auth_session_flg. eq N) %then %do;
      /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
      options &suppressLogOptions.;
      
      /* Get the Service Ticket. Returns service ticket in the macro variable __irm_rest_service_ticket__ */
      %irm_rest_service_ticket(
                url          = &rest_url.
              , appcontext   = &server.
              , irm_username = &username.
              , irm_password = &passwd.
              , debug        = &debug.
              , logOptions   = &oldLogOptions.
       );
   %end;
   
   /* Restore original log options. */
   options &oldLogOptions.;

   %if (%bquote(&fa_tgt_path.) eq %str()) %then
      %let fa_tgt_path = &fa_src_path.;
   
   /* Make the REST call. */
   filename json_in   temp;
   filename json_out  temp;
   
   /* Build the JSON body for the POST request. Need to pass as payload:
      sourcepath, forceFlag */
   filename json_in temp;
   %put NOTE: Creating JSON payload for a new FA;
   proc json out = json_in pretty;
      write open o ; 
         write values "sourcePath" "&fa_src_path.";
         write values "forceFlag" "&force_flg.";
      write close;
   run;  

   /* Send the request */
   proc http
      in        = json_in
      out       = json_out
      method    = 'put'
      ct        = 'application/json'
      %if(&auth_session_flg. eq Y) %then %do;
         /* No credentials -> use Token authentication (this is only allowed if running the code from an authenticated SAS Session) */
         url = "%str(&rest_url.)"
         http_tokenauth
      %end;
      %else %do;
         /* Use the ticket */
         url  = "&rest_url?ticket=&__irm_rest_service_ticket__." 
      %end;
      /* Clear connection cache */
      clear_cache
      ;
   run;

   /* Collect the response from the server and print it to the log */
   data _null_;
      infile json_out;
      input;
      %if (%upcase(&debug.) eq TRUE) %then %do;       
         put 'Response from server:';
      %end;
      put _infile_;
   run;
   
   /* Clear filerefs */   
   filename json_out  clear;
   filename json_in   clear;

%mend;
